@@ -18,7 +18,6 @@
 open Misc
 open Asttypes
 open Primitive
-open Types
 open Lambda
 open Switch
 open Instruct
@@ -34,20 +33,53 @@ let new_label () =
 (**** Operations on compilation environments. ****)
 
 let empty_env =
-  { ce_stack = Ident.empty; ce_heap = Ident.empty; ce_rec = Ident.empty }
+  { ce_stack = Ident.empty; ce_closure = Not_in_closure }
 
 (* Add a stack-allocated variable *)
 
 let add_var id pos env =
   { ce_stack = Ident.add id pos env.ce_stack;
-    ce_heap = env.ce_heap;
-    ce_rec = env.ce_rec }
+    ce_closure = env.ce_closure }
 
 let rec add_vars idlist pos env =
   match idlist with
     [] -> env
   | id :: rem -> add_vars rem (pos + 1) (add_var id pos env)
 
+(* Compute the closure environment *)
+
+let rec add_positions entries pos_to_entry ~pos ~delta = function
+  | [] -> entries, pos
+  | id :: rem ->
+    let entries =
+      Ident.add id (pos_to_entry pos) entries
+    in
+    add_positions entries pos_to_entry ~pos:(pos + delta) ~delta rem
+
+type function_definition =
+  | Single_non_recursive
+  | Multiple_recursive of Ident.t list
+
+let closure_entries fun_defs fvs =
+  let funct_entries, pos_end_functs =
+    match fun_defs with
+    | Single_non_recursive ->
+      (* No need to store the function in the environment, but we still need to
+         reserve a slot in the closure block *)
+      Ident.empty, 3
+    | Multiple_recursive functs ->
+      add_positions Ident.empty (fun pos -> Function pos) ~pos:0 ~delta:3 functs
+  in
+  (* Note: [pos_end_functs] is the position where we would store the next
+     function if there was one, and points after an eventual infix tag.
+     Since that was the last function, we don't need the last infix tag
+     and start storing free variables at [pos_end_functs - 1]. *)
+  let all_entries, _end_pos =
+    add_positions funct_entries (fun pos -> Free_variable pos)
+      ~pos:(pos_end_functs - 1) ~delta:1 fvs
+  in
+  all_entries
+
 (**** Examination of the continuation ****)
 
 (* Return a label to the beginning of the given continuation.
@@ -157,76 +189,6 @@ let rec push_dummies n k = match n with
 | _ -> Kconst const_unit::Kpush::push_dummies (n-1) k
 
 
-(**** Auxiliary for compiling "let rec" ****)
-
-type rhs_kind =
-  | RHS_block of int
-  | RHS_infix of { blocksize : int; offset : int }
-  | RHS_floatblock of int
-  | RHS_nonrec
-  | RHS_function of int * int
-
-let rec check_recordwith_updates id e =
-  match e with
-  | Lsequence (Lprim ((Psetfield _ | Psetfloatfield _), [Lvar id2; _], _), cont)
-      -> id2 = id && check_recordwith_updates id cont
-  | Lvar id2 -> id2 = id
-  | _ -> false
-
-let rec size_of_lambda env = function
-  | Lvar id ->
-      begin try Ident.find_same id env with Not_found -> RHS_nonrec end
-  | Lfunction{params} as funct ->
-      RHS_function (2 + Ident.Set.cardinal(free_variables funct),
-                    List.length params)
-  | Llet (Strict, _k, id, Lprim (Pduprecord (kind, size), _, _), body)
-    when check_recordwith_updates id body ->
-      begin match kind with
-      | Record_regular | Record_inlined _ -> RHS_block size
-      | Record_unboxed _ -> assert false
-      | Record_float -> RHS_floatblock size
-      | Record_extension _ -> RHS_block (size + 1)
-      end
-  | Llet(_str, _k, id, arg, body) ->
-      size_of_lambda (Ident.add id (size_of_lambda env arg) env) body
-  (* See the Lletrec case of comp_expr *)
-  | Lletrec(bindings, body) when
-      List.for_all (function (_, Lfunction _) -> true | _ -> false) bindings ->
-      (* let rec of functions *)
-      let fv =
-        Ident.Set.elements (free_variables (Lletrec(bindings, lambda_unit))) in
-      (* See Instruct(CLOSUREREC) in interp.c *)
-      let blocksize = List.length bindings * 3 - 1 + List.length fv in
-      let offsets = List.mapi (fun i (id, _e) -> (id, i * 3)) bindings in
-      let env = List.fold_right (fun (id, offset) env ->
-        Ident.add id (RHS_infix { blocksize; offset }) env) offsets env in
-      size_of_lambda env body
-  | Lletrec(bindings, body) ->
-      let env = List.fold_right
-        (fun (id, e) env -> Ident.add id (size_of_lambda env e) env)
-        bindings env
-      in
-      size_of_lambda env body
-  | Lprim(Pmakeblock _, args, _) -> RHS_block (List.length args)
-  | Lprim (Pmakearray ((Paddrarray|Pintarray), _), args, _) ->
-      RHS_block (List.length args)
-  | Lprim (Pmakearray (Pfloatarray, _), args, _) ->
-      RHS_floatblock (List.length args)
-  | Lprim (Pmakearray (Pgenarray, _), _, _) ->
-     (* Pgenarray is excluded from recursive bindings by the
-        check in Translcore.check_recursive_lambda *)
-      RHS_nonrec
-  | Lprim (Pduprecord ((Record_regular | Record_inlined _), size), _, _) ->
-      RHS_block size
-  | Lprim (Pduprecord (Record_unboxed _, _), _, _) ->
-      assert false
-  | Lprim (Pduprecord (Record_extension _, size), _, _) ->
-      RHS_block (size + 1)
-  | Lprim (Pduprecord (Record_float, size), _, _) -> RHS_floatblock size
-  | Levent (lam, _) -> size_of_lambda env lam
-  | Lsequence (_lam, lam') -> size_of_lambda env lam'
-  | _ -> RHS_nonrec
-
 (**** Merging consecutive events ****)
 
 let copy_event ev kind info repr =
@@ -370,9 +332,8 @@ type function_to_compile =
   { params: Ident.t list;               (* function parameters *)
     body: lambda;                       (* the function body *)
     label: label;                       (* the label of the function entry *)
-    free_vars: Ident.t list;            (* free variables of the function *)
-    num_defs: int;            (* number of mutually recursive definitions *)
-    rec_vars: Ident.t list;             (* mutually recursive fn names *)
+    entries: closure_entry Ident.tbl;   (* the offsets for the free variables
+                                           and mutually recursive functions *)
     rec_pos: int }                      (* rank in recursive definition *)
 
 let functions_to_compile  = (Stack.create () : function_to_compile Stack.t)
@@ -570,15 +531,18 @@ let rec comp_expr stack_info env exp sz cont =
         let pos = Ident.find_same id env.ce_stack in
         Kacc(sz - pos) :: cont
       with Not_found ->
-      try
-        let pos = Ident.find_same id env.ce_heap in
-        Kenvacc(pos) :: cont
-      with Not_found ->
-      try
-        let ofs = Ident.find_same id env.ce_rec in
-        Koffsetclosure(ofs) :: cont
-      with Not_found ->
+      let not_found () =
         fatal_error ("Bytegen.comp_expr: var " ^ Ident.unique_name id)
+      in
+      match env.ce_closure with
+      | Not_in_closure -> not_found ()
+      | In_closure { entries; env_pos } ->
+        match Ident.find_same id entries with
+        | Free_variable pos ->
+          Kenvacc(pos - env_pos) :: cont
+        | Function pos ->
+          Koffsetclosure(pos - env_pos) :: cont
+        | exception Not_found -> not_found ()
       end
   | Lconst cst ->
       Kconst cst :: cont
@@ -627,9 +591,10 @@ let rec comp_expr stack_info env exp sz cont =
       let cont = add_pseudo_event loc !compunit_name cont in
       let lbl = new_label() in
       let fv = Ident.Set.elements(free_variables exp) in
+      let entries = closure_entries Single_non_recursive fv in
       let to_compile =
         { params = List.map fst params; body = body; label = lbl;
-          free_vars = fv; num_defs = 1; rec_vars = []; rec_pos = 0 } in
+          entries = entries; rec_pos = 0 } in
       Stack.push to_compile functions_to_compile;
       comp_args stack_info env (List.map (fun n -> Lvar n) fv) sz
         (Kclosure(lbl, List.length fv) :: cont)
@@ -640,80 +605,28 @@ let rec comp_expr stack_info env exp sz cont =
           (add_pop 1 cont))
   | Lletrec(decl, body) ->
       let ndecl = List.length decl in
-      if List.for_all (function (_, Lfunction _) -> true | _ -> false)
-                      decl then begin
-        (* let rec of functions *)
-        let fv =
-          Ident.Set.elements (free_variables (Lletrec(decl, lambda_unit))) in
-        let rec_idents = List.map (fun (id, _lam) -> id) decl in
-        let rec comp_fun pos = function
-            [] -> []
-          | (_id, Lfunction{params; body}) :: rem ->
-              let lbl = new_label() in
-              let to_compile =
-                { params = List.map fst params; body = body; label = lbl;
-                  free_vars = fv; num_defs = ndecl; rec_vars = rec_idents;
-                  rec_pos = pos} in
-              Stack.push to_compile functions_to_compile;
-              lbl :: comp_fun (pos + 1) rem
-          | _ -> assert false in
-        let lbls = comp_fun 0 decl in
-        comp_args stack_info env (List.map (fun n -> Lvar n) fv) sz
-          (Kclosurerec(lbls, List.length fv) ::
-           (comp_expr stack_info
-              (add_vars rec_idents (sz+1) env) body (sz + ndecl)
-              (add_pop ndecl cont)))
-      end else begin
-        let decl_size =
-          List.map (fun (id, exp) -> (id, exp, size_of_lambda Ident.empty exp))
-            decl in
-        let rec comp_init new_env sz = function
-          | [] -> comp_nonrec new_env sz ndecl decl_size
-          | (id, _exp, RHS_floatblock blocksize) :: rem ->
-              Kconst(Const_base(Const_int blocksize)) ::
-              Kccall("caml_alloc_dummy_float", 1) :: Kpush ::
-              comp_init (add_var id (sz+1) new_env) (sz+1) rem
-          | (id, _exp, RHS_block blocksize) :: rem ->
-              Kconst(Const_base(Const_int blocksize)) ::
-              Kccall("caml_alloc_dummy", 1) :: Kpush ::
-              comp_init (add_var id (sz+1) new_env) (sz+1) rem
-          | (id, _exp, RHS_infix { blocksize; offset }) :: rem ->
-              Kconst(Const_base(Const_int offset)) ::
-              Kpush ::
-              Kconst(Const_base(Const_int blocksize)) ::
-              Kccall("caml_alloc_dummy_infix", 2) :: Kpush ::
-              comp_init (add_var id (sz+1) new_env) (sz+1) rem
-          | (id, _exp, RHS_function (blocksize,arity)) :: rem ->
-              Kconst(Const_base(Const_int arity)) ::
-              Kpush ::
-              Kconst(Const_base(Const_int blocksize)) ::
-              Kccall("caml_alloc_dummy_function", 2) :: Kpush ::
-              comp_init (add_var id (sz+1) new_env) (sz+1) rem
-          | (id, _exp, RHS_nonrec) :: rem ->
-              Kconst(Const_base(Const_int 0)) :: Kpush ::
-              comp_init (add_var id (sz+1) new_env) (sz+1) rem
-        and comp_nonrec new_env sz i = function
-          | [] -> comp_rec new_env sz ndecl decl_size
-          | (_id, _exp, (RHS_block _ | RHS_infix _ |
-                         RHS_floatblock _ | RHS_function _))
-            :: rem ->
-              comp_nonrec new_env sz (i-1) rem
-          | (_id, exp, RHS_nonrec) :: rem ->
-              comp_expr stack_info new_env exp sz
-                (Kassign (i-1) :: comp_nonrec new_env sz (i-1) rem)
-        and comp_rec new_env sz i = function
-          | [] -> comp_expr stack_info new_env body sz (add_pop ndecl cont)
-          | (_id, exp, (RHS_block _ | RHS_infix _ |
-                        RHS_floatblock _ | RHS_function _))
-            :: rem ->
-              comp_expr stack_info new_env exp sz
-                (Kpush :: Kacc i :: Kccall("caml_update_dummy", 2) ::
-                 comp_rec new_env sz (i-1) rem)
-          | (_id, _exp, RHS_nonrec) :: rem ->
-              comp_rec new_env sz (i-1) rem
-        in
-        comp_init env sz decl_size
-      end
+      let fv =
+        Ident.Set.elements (free_variables (Lletrec(decl, lambda_unit))) in
+      let rec_idents = List.map (fun { id } -> id) decl in
+      let entries =
+        closure_entries (Multiple_recursive rec_idents) fv
+      in
+      let rec comp_fun pos = function
+          [] -> []
+        | { def = {params; body} } :: rem ->
+            let lbl = new_label() in
+            let to_compile =
+              { params = List.map fst params; body = body; label = lbl;
+                entries = entries; rec_pos = pos} in
+            Stack.push to_compile functions_to_compile;
+            lbl :: comp_fun (pos + 1) rem
+      in
+      let lbls = comp_fun 0 decl in
+      comp_args stack_info env (List.map (fun n -> Lvar n) fv) sz
+        (Kclosurerec(lbls, List.length fv) ::
+         (comp_expr stack_info
+            (add_vars rec_idents (sz+1) env) body (sz + ndecl)
+            (add_pop ndecl cont)))
   | Lprim(Popaque, [arg], _) ->
       comp_expr stack_info env arg sz cont
   | Lprim(Pignore, [arg], _) ->
@@ -783,16 +696,34 @@ let rec comp_expr stack_info env exp sz cont =
                  (Kmakeblock(List.length args, 0) ::
                   Kccall("caml_make_array", 1) :: cont)
       end
-  | Lprim((Presume|Prunstack), args, _) ->
+  | Lprim(Presume, args, _) ->
       let nargs = List.length args - 1 in
-      assert (nargs = 2);
-      (* Resume itself only pushes 3 words, but perform adds another *)
-      check_stack stack_info (sz + 4);
-      if is_tailcall cont then
+      assert (nargs = 3);
+      if is_tailcall cont then begin
+        (* Resumeterm itself only pushes 2 words, but perform adds another *)
+        check_stack stack_info 3;
         comp_args stack_info env args sz
           (Kresumeterm(sz + nargs) :: discard_dead_code cont)
-      else
+      end else begin
+        (* Resume itself only pushes 2 words, but perform adds another *)
+        check_stack stack_info (sz + nargs + 3);
         comp_args stack_info env args sz (Kresume :: cont)
+      end
+  | Lprim(Prunstack, args, _) ->
+      let nargs = List.length args in
+      assert (nargs = 3);
+      if is_tailcall cont then begin
+        (* Resumeterm itself only pushes 2 words, but perform adds another *)
+        check_stack stack_info 3;
+        Kconst const_unit :: Kpush ::
+          comp_args stack_info env args (sz + 1)
+          (Kresumeterm(sz + nargs) :: discard_dead_code cont)
+      end else begin
+        (* Resume itself only pushes 2 words, but perform adds another *)
+        check_stack stack_info (sz + nargs + 3);
+        Kconst const_unit :: Kpush ::
+          comp_args stack_info env args (sz + 1) (Kresume :: cont)
+      end
   | Lprim(Preperform, args, _) ->
       let nargs = List.length args - 1 in
       assert (nargs = 2);
@@ -1119,13 +1050,15 @@ let comp_block env exp sz cont =
 
 let comp_function tc cont =
   let arity = List.length tc.params in
-  let rec positions pos delta = function
-      [] -> Ident.empty
-    | id :: rem -> Ident.add id pos (positions (pos + delta) delta rem) in
+  let ce_stack, _last_pos =
+    add_positions Ident.empty Fun.id ~pos:arity ~delta:(-1) tc.params
+  in
   let env =
-    { ce_stack = positions arity (-1) tc.params;
-      ce_heap = positions (3 * (tc.num_defs - tc.rec_pos) - 1) 1 tc.free_vars;
-      ce_rec = positions (-3 * tc.rec_pos) 3 tc.rec_vars } in
+    { ce_stack;
+      ce_closure =
+        In_closure { entries = tc.entries; env_pos = 3 * tc.rec_pos }
+    }
+  in
   let cont =
     comp_block env tc.body arity (Kreturn arity :: cont) in
   if arity > 1 then
@@ -1151,20 +1084,23 @@ let reset () =
   compunit_name := "";
   Stack.clear functions_to_compile
 
-let compile_implementation modulename expr =
+let compile_gen ?modulename ~init_stack expr =
   reset ();
-  compunit_name := modulename;
+  begin match modulename with
+  | Some name -> compunit_name := name
+  | None -> ()
+  end;
   Fun.protect ~finally:reset (fun () ->
-  let init_code = comp_block empty_env expr 0 [] in
+  let init_code = comp_block empty_env expr init_stack [] in
   if Stack.length functions_to_compile > 0 then begin
     let lbl_init = new_label() in
-    Kbranch lbl_init :: comp_remainder (Klabel lbl_init :: init_code)
+    (Kbranch lbl_init :: comp_remainder (Klabel lbl_init :: init_code)),
+    false
   end else
-    init_code)
+    init_code, true)
+
+let compile_implementation modulename expr =
+  fst (compile_gen ~modulename ~init_stack:0 expr)
 
 let compile_phrase expr =
-  reset ();
-  Fun.protect ~finally:reset (fun () ->
-  let init_code = comp_block empty_env expr 1 [Kreturn 1] in
-  let fun_code = comp_remainder [] in
-  (init_code, fun_code))
+  compile_gen ~init_stack:1 expr
