@@ -447,7 +447,7 @@ module Analyser =
 
     let get_type_kind env name_comment_list type_kind =
       match type_kind with
-        Types.Type_abstract ->
+        Types.Type_abstract _ ->
           Odoc_type.Type_abstract
       | Types.Type_variant (l,_) ->
           let f {Types.cd_id=constructor_name;cd_args;cd_res=ret_type;cd_attributes} =
@@ -1885,9 +1885,7 @@ module Analyser =
         (ast : Parsetree.signature) (signat : Types.signature) =
       prepare_file source_file input_file;
       (* We create the t_module for this file. *)
-      let mod_name = String.capitalize_ascii
-          (Filename.basename (try Filename.chop_extension source_file with _ -> source_file))
-      in
+      let mod_name = Unit_info.modname_from_source source_file in
       let len, info_opt = preamble !file_name !file
           (fun x -> x.Parsetree.psig_loc) ast in
       let info_opt = analyze_toplevel_alerts info_opt ast in
