@@ -20,10 +20,14 @@
 
 #define _GNU_SOURCE  /* For sched.h CPU_ZERO(3) and CPU_COUNT(3) */
 #include "caml/config.h"
+#include <stdbool.h>
 #include <stdio.h>
+#ifdef HAS_UNISTD
 #include <unistd.h>
+#endif
 #include <pthread.h>
 #include <string.h>
+#include <assert.h>
 #ifdef HAS_GNU_GETAFFINITY_NP
 #include <sched.h>
 #ifdef HAS_PTHREAD_NP_H
@@ -36,6 +40,8 @@
 typedef cpuset_t cpu_set_t;
 #endif
 #ifdef _WIN32
+#define WIN32_LEAN_AND_MEAN
+#include <windows.h>
 #include <sysinfoapi.h>
 #endif
 #include "caml/alloc.h"
@@ -54,6 +60,7 @@ typedef cpuset_t cpu_set_t;
 #include "caml/intext.h"
 #include "caml/major_gc.h"
 #include "caml/minor_gc.h"
+#include "caml/memprof.h"
 #include "caml/misc.h"
 #include "caml/memory.h"
 #include "caml/osdeps.h"
@@ -64,6 +71,13 @@ typedef cpuset_t cpu_set_t;
 #include "caml/sync.h"
 #include "caml/weak.h"
 
+/* Check that the domain_state structure was laid out without padding,
+   since the runtime assumes this in computing offsets */
+static_assert(
+    offsetof(caml_domain_state, LAST_DOMAIN_STATE_MEMBER) ==
+    (Domain_state_num_fields - 1) * 8,
+    "");
+
 /* The runtime can run stop-the-world (STW) sections, during which all
    active domains run the same callback in parallel (with a barrier
    mechanism to synchronize within the callback).
@@ -141,7 +155,10 @@ typedef cpuset_t cpu_set_t;
 
 /* control of STW interrupts */
 struct interruptor {
-  atomic_uintnat* interrupt_word;
+  /* The outermost atomic is for synchronization with
+     caml_interrupt_all_signal_safe. The innermost atomic is also for
+     cross-domain communication.*/
+  _Atomic(atomic_uintnat *) interrupt_word;
   caml_plat_mutex lock;
   caml_plat_cond cond;
 
@@ -205,7 +222,7 @@ static caml_plat_mutex all_domains_lock = CAML_PLAT_MUTEX_INITIALIZER;
 static caml_plat_cond all_domains_cond =
     CAML_PLAT_COND_INITIALIZER(&all_domains_lock);
 static atomic_uintnat /* dom_internal* */ stw_leader = 0;
-static struct dom_internal all_domains[Max_domains];
+static dom_internal all_domains[Max_domains];
 
 CAMLexport atomic_uintnat caml_num_domains_running;
 
@@ -230,7 +247,7 @@ uintnat caml_minor_heap_max_wsz;
 
 CAMLexport uintnat caml_minor_heaps_start;
 CAMLexport uintnat caml_minor_heaps_end;
-static __thread dom_internal* domain_self;
+static CAMLthread_local dom_internal* domain_self;
 
 /*
  * This structure is protected by all_domains_lock
@@ -245,19 +262,22 @@ static struct {
   { 0 }
 };
 
-static void add_to_stw_domains(dom_internal* dom) {
-  int i;
+static void add_next_to_stw_domains(void)
+{
   CAMLassert(stw_domains.participating_domains < Max_domains);
-  for(i=stw_domains.participating_domains; stw_domains.domains[i]!=dom; ++i) {
-    CAMLassert(i<Max_domains);
-  }
-
-  /* swap passed domain with domain at stw_domains.participating_domains */
-  dom = stw_domains.domains[stw_domains.participating_domains];
-  stw_domains.domains[stw_domains.participating_domains] =
-      stw_domains.domains[i];
-  stw_domains.domains[i] = dom;
   stw_domains.participating_domains++;
+#ifdef DEBUG
+  /* Enforce here the invariant for early-exit in
+     [caml_interrupt_all_signal_safe], because the latter must be
+     async-signal-safe and one cannot CAMLassert inside it. */
+  bool prev_has_interrupt_word = true;
+  for (int i = 0; i < Max_domains; i++) {
+    bool has_interrupt_word = all_domains[i].interruptor.interrupt_word != NULL;
+    if (i < stw_domains.participating_domains) CAMLassert(has_interrupt_word);
+    if (!prev_has_interrupt_word) CAMLassert(!has_interrupt_word);
+    prev_has_interrupt_word = has_interrupt_word;
+  }
+#endif
 }
 
 static void remove_from_stw_domains(dom_internal* dom) {
@@ -282,7 +302,7 @@ static dom_internal* next_free_domain(void) {
   return stw_domains.domains[stw_domains.participating_domains];
 }
 
-CAMLexport __thread caml_domain_state* caml_state;
+CAMLexport CAMLthread_local caml_domain_state* caml_state;
 
 #ifndef HAS_FULL_THREAD_VARIABLES
 /* Export a getter for caml_state, to be used in DLLs */
@@ -294,7 +314,13 @@ CAMLexport caml_domain_state* caml_get_domain_state(void)
 
 Caml_inline void interrupt_domain(struct interruptor* s)
 {
-  atomic_store_release(s->interrupt_word, (uintnat)(-1));
+  atomic_uintnat * interrupt_word = atomic_load_relaxed(&s->interrupt_word);
+  atomic_store_release(interrupt_word, UINTNAT_MAX);
+}
+
+Caml_inline void interrupt_domain_local(caml_domain_state* dom_st)
+{
+  atomic_store_relaxed(&dom_st->young_limit, UINTNAT_MAX);
 }
 
 int caml_incoming_interrupts_queued(void)
@@ -495,6 +521,7 @@ static int allocate_minor_heap(asize_t wsize) {
    * major slice is scheduled. */
   domain_state->young_trigger = domain_state->young_start
          + (domain_state->young_end - domain_state->young_start) / 2;
+  caml_memprof_renew_minor_sample(domain_state);
   caml_reset_young_limit(domain_state);
 
   check_minor_heap();
@@ -532,7 +559,9 @@ static uintnat fresh_domain_unique_id(void) {
 }
 
 /* must be run on the domain's thread */
-static void domain_create(uintnat initial_minor_heap_wsize) {
+static void domain_create(uintnat initial_minor_heap_wsize,
+                          caml_domain_state *parent)
+{
   dom_internal* d = 0;
   caml_domain_state* domain_state;
   struct interruptor* s;
@@ -560,8 +589,6 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
   CAMLassert(!s->running);
   CAMLassert(!s->interrupt_pending);
 
-  domain_self = d;
-
   /* If the chosen domain slot has not been previously used, allocate a fresh
      domain state. Otherwise, reuse it.
 
@@ -583,18 +610,35 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
     domain_state = d->state;
   }
 
-  caml_state = domain_state;
-
-  s->unique_id = fresh_domain_unique_id();
-  s->interrupt_word = &domain_state->young_limit;
-  s->running = 1;
-  atomic_fetch_add(&caml_num_domains_running, 1);
+  /* Note: until we take d->domain_lock, the domain_state may still be
+   * shared with a domain which is terminating (see
+   * domain_terminate). */
 
   caml_plat_lock(&d->domain_lock);
 
+  /* Set domain_self if we have successfully allocated the
+   * caml_domain_state. Otherwise domain_self will be NULL and it's up
+   * to the caller to deal with that. */
+
+  domain_self = d;
+  caml_state = domain_state;
+
+  domain_state->young_limit = 0;
+  /* Synchronized with [caml_interrupt_all_signal_safe], so that the
+     initializing write of young_limit happens before any
+     interrupt. */
+  atomic_store_explicit(&s->interrupt_word, &domain_state->young_limit,
+                        memory_order_release);
+
+  /* Tell memprof system about the new domain before either (a) new
+   * domain can allocate anything or (b) parent domain can go away. */
+  caml_memprof_new_domain(parent, domain_state);
+  if (!domain_state->memprof) {
+    goto init_memprof_failure;
+  }
+
   domain_state->id = d->id;
-  domain_state->unique_id = d->interruptor.unique_id;
-  CAMLassert(!d->interruptor.interrupt_pending);
+  CAMLassert(!s->interrupt_pending);
 
   domain_state->extra_heap_resources = 0.0;
   domain_state->extra_heap_resources_minor = 0.0;
@@ -647,6 +691,14 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
     goto alloc_main_stack_failure;
   }
 
+  /* No remaining failure cases: domain creation is going to succeed,
+   * so we can update globally-visible state without needing to unwind
+   * it. */
+  s->unique_id = fresh_domain_unique_id();
+  domain_state->unique_id = s->unique_id;
+  s->running = 1;
+  atomic_fetch_add(&caml_num_domains_running, 1);
+
   domain_state->c_stack = NULL;
   domain_state->exn_handler = NULL;
 
@@ -687,7 +739,7 @@ static void domain_create(uintnat initial_minor_heap_wsize) {
 #endif
 
   caml_reset_young_limit(domain_state);
-  add_to_stw_domains(domain_self);
+  add_next_to_stw_domains();
   goto domain_init_complete;
 
 alloc_main_stack_failure:
@@ -701,6 +753,8 @@ init_shared_heap_failure:
   caml_free_minor_tables(domain_state->minor_tables);
   domain_state->minor_tables = NULL;
 alloc_minor_tables_failure:
+  caml_memprof_delete_domain(domain_state);
+init_memprof_failure:
   domain_self = NULL;
 
 
@@ -713,6 +767,19 @@ CAMLexport void caml_reset_domain_lock(void)
 {
   dom_internal* self = domain_self;
   // This is only used to reset the domain_lock state on fork.
+  /* FIXME: initializing an already-initialized mutex and cond
+     variable is UB (especially mutexes that are locked).
+
+     * On systhreads, this is best-effort but at least the error
+       conditions should be checked and reported.
+
+     * If there is only one thread, it is sensible to fork but the
+       mutex should still not be initialized while locked. On Linux it
+       seems that the mutex remains valid and locked
+       (https://man7.org/linux/man-pages/man2/fork.2.html). For
+       portability on POSIX the lock should be released and destroyed
+       prior to calling fork and then init afterwards in both parent
+       and child. */
   caml_plat_mutex_init(&self->domain_lock);
   caml_plat_cond_init(&self->domain_cond, &self->domain_lock);
 
@@ -721,7 +788,7 @@ CAMLexport void caml_reset_domain_lock(void)
 
 /* minor heap initialization and resizing */
 
-static void reserve_minor_heaps(void) {
+static void reserve_minor_heaps_from_stw_single(void) {
   void* heaps_base;
   uintnat minor_heap_reservation_bsize;
   uintnat minor_heap_max_bsz;
@@ -733,8 +800,7 @@ static void reserve_minor_heaps(void) {
   minor_heap_reservation_bsize = minor_heap_max_bsz * Max_domains;
 
   /* reserve memory space for minor heaps */
-  heaps_base = caml_mem_map(minor_heap_reservation_bsize, caml_plat_pagesize,
-                1 /* reserve_only */);
+  heaps_base = caml_mem_map(minor_heap_reservation_bsize, 1 /* reserve_only */);
   if (heaps_base == NULL)
     caml_fatal_error("Not enough heap memory to reserve minor heaps");
 
@@ -758,7 +824,7 @@ static void reserve_minor_heaps(void) {
   }
 }
 
-static void unreserve_minor_heaps(void) {
+static void unreserve_minor_heaps_from_stw_single(void) {
   uintnat size;
 
   caml_gc_log("unreserve_minor_heaps");
@@ -810,16 +876,16 @@ static void stw_resize_minor_heap_reservation(caml_domain_state* domain,
     caml_gc_log("stw_resize_minor_heap_reservation: "
                 "unreserve_minor_heaps");
 
-    unreserve_minor_heaps();
+    unreserve_minor_heaps_from_stw_single();
     /* new_minor_wsz is page-aligned because caml_norm_minor_heap_size has
        been called to normalize it earlier.
     */
     caml_minor_heap_max_wsz = new_minor_wsz;
     caml_gc_log("stw_resize_minor_heap_reservation: reserve_minor_heaps");
-    reserve_minor_heaps();
-    /* The call to [reserve_minor_heaps] makes a new reservation,
-       and it also updates the reservation boundaries of each domain
-       by mutating its [minor_heap_area_start{,_end}] variables.
+    reserve_minor_heaps_from_stw_single();
+    /* The call to [reserve_minor_heaps_from_stw_single] makes a new
+       reservation, and it also updates the reservation boundaries of each
+       domain by mutating its [minor_heap_area_start{,_end}] variables.
 
        These variables are synchronized by the fact that we are inside
        a STW section: no other domains are running in parallel, and
@@ -855,7 +921,9 @@ void caml_update_minor_heap_max(uintnat requested_wsz) {
 void caml_init_domains(uintnat minor_heap_wsz) {
   int i;
 
-  reserve_minor_heaps();
+  reserve_minor_heaps_from_stw_single();
+  /* stw_single: mutators and domains have not started yet. */
+
   for (i = 0; i < Max_domains; i++) {
     struct dom_internal* dom = &all_domains[i];
 
@@ -863,7 +931,7 @@ void caml_init_domains(uintnat minor_heap_wsz) {
 
     dom->id = i;
 
-    dom->interruptor.interrupt_word = 0;
+    dom->interruptor.interrupt_word = NULL;
     caml_plat_mutex_init(&dom->interruptor.lock);
     caml_plat_cond_init(&dom->interruptor.cond,
                         &dom->interruptor.lock);
@@ -878,7 +946,7 @@ void caml_init_domains(uintnat minor_heap_wsz) {
     dom->backup_thread_msg = BT_INIT;
   }
 
-  domain_create(minor_heap_wsz);
+  domain_create(minor_heap_wsz, NULL);
   if (!domain_self) caml_fatal_error("Failed to create main domain");
   CAMLassert (domain_self->state->unique_id == 0);
 
@@ -895,21 +963,27 @@ enum domain_status { Dom_starting, Dom_started, Dom_failed };
 
 struct domain_ml_values {
   value callback;
-  value term_mutex;
+  value term_sync;
 };
 
+/* stdlib/domain.ml */
+#define Term_state(sync) (&Field(sync, 0))
+#define Term_mutex(sync) (&Field(sync, 1))
+#define Term_condition(sync) (&Field(sync, 2))
+
 static void init_domain_ml_values(struct domain_ml_values* ml_values,
-                                  value callback, value term_mutex)
+                                  value callback, value term_sync)
 {
   ml_values->callback = callback;
-  ml_values->term_mutex = term_mutex;
+  ml_values->term_sync = term_sync;
   caml_register_generational_global_root(&ml_values->callback);
-  caml_register_generational_global_root(&ml_values->term_mutex);
+  caml_register_generational_global_root(&ml_values->term_sync);
 }
 
-static void free_domain_ml_values(struct domain_ml_values* ml_values) {
+static void free_domain_ml_values(struct domain_ml_values* ml_values)
+{
   caml_remove_generational_global_root(&ml_values->callback);
-  caml_remove_generational_global_root(&ml_values->term_mutex);
+  caml_remove_generational_global_root(&ml_values->term_sync);
   caml_stat_free(ml_values);
 }
 
@@ -919,16 +993,12 @@ static void free_domain_ml_values(struct domain_ml_values* ml_values) {
    parameters returned to the parent by the child.
 */
 struct domain_startup_params {
-  struct interruptor* parent; /* in */
+  dom_internal *parent; /* in */
   enum domain_status status; /* in+out:
                                 parent and child synchronize on this value. */
   struct domain_ml_values* ml_values; /* in */
   dom_internal* newdom; /* out */
   uintnat unique_id; /* out */
-#ifndef _WIN32
-  /* signal mask to set after it is safe to do so */
-  sigset_t* mask; /* in */
-#endif
 };
 
 static void* backup_thread_func(void* v)
@@ -1060,62 +1130,109 @@ CAMLexport _Atomic caml_timing_hook caml_domain_terminated_hook =
 
 static void domain_terminate(void);
 
+static value make_finished(value res_or_exn)
+{
+  CAMLparam0();
+  CAMLlocal1(res);
+  if (Is_exception_result(res_or_exn)) {
+    res = Extract_exception(res_or_exn);
+    /* [Error res] */
+    res = caml_alloc_1(1, res);
+  } else {
+    /* [Ok res_of_exn] */
+    res = caml_alloc_1(0, res_or_exn);
+  }
+  /* [Finished res] */
+  res = caml_alloc_1(0, res);
+  CAMLreturn(res);
+}
+
+static void sync_result(value term_sync, value res)
+{
+  CAMLparam2(term_sync, res);
+  /* Synchronize with joining domains. We call [caml_ml_mutex_lock]
+     because the systhreads are still running on this domain. We
+     assume this does not fail the exception it would raise at this
+     point would be bad for us. */
+  caml_ml_mutex_lock(*Term_mutex(term_sync));
+
+  /* Store result */
+  volatile value *state = Term_state(term_sync);
+  CAMLassert(!Is_block(*state));
+  caml_modify(state, res);
+
+  /* Signal all the waiting domains to be woken up */
+  caml_ml_condition_broadcast(*Term_condition(term_sync));
+
+  /* The mutex is unlocked in the runtime after the cleanup
+     functions are finished. */
+  CAMLreturn0;
+}
+
 static void* domain_thread_func(void* v)
 {
   struct domain_startup_params* p = v;
   struct domain_ml_values *ml_values = p->ml_values;
-#ifndef _WIN32
-  sigset_t mask = *(p->mask);
-  void * signal_stack;
 
-  signal_stack = caml_init_signal_stack();
+#ifndef _WIN32
+  void * signal_stack = caml_init_signal_stack();
   if (signal_stack == NULL) {
     caml_fatal_error("Failed to create domain: signal stack");
   }
 #endif
 
-  domain_create(caml_params->init_minor_heap_wsz);
+  domain_create(caml_params->init_minor_heap_wsz, p->parent->state);
+
   /* this domain is now part of the STW participant set */
   p->newdom = domain_self;
 
   /* handshake with the parent domain */
-  caml_plat_lock(&p->parent->lock);
+  caml_plat_lock(&p->parent->interruptor.lock);
   if (domain_self) {
     p->status = Dom_started;
     p->unique_id = domain_self->interruptor.unique_id;
   } else {
     p->status = Dom_failed;
   }
-  caml_plat_broadcast(&p->parent->cond);
-  caml_plat_unlock(&p->parent->lock);
+  caml_plat_broadcast(&p->parent->interruptor.cond);
+  caml_plat_unlock(&p->parent->interruptor.lock);
   /* Cannot access p below here. */
 
   if (domain_self) {
     install_backup_thread(domain_self);
 
-#ifndef _WIN32
-    /* It is now safe for us to handle signals */
-    pthread_sigmask(SIG_SETMASK, &mask, NULL);
-#endif
-
     caml_gc_log("Domain starting (unique_id = %"ARCH_INTNAT_PRINTF_FORMAT"u)",
                 domain_self->interruptor.unique_id);
     CAML_EV_LIFECYCLE(EV_DOMAIN_SPAWN, getpid());
+    /* FIXME: ignoring errors during domain initialization is unsafe
+       and/or can deadlock. */
     caml_domain_initialize_hook();
-    caml_callback(ml_values->callback, Val_unit);
+
+    /* release callback early;
+       see the [note about callbacks and GC] in callback.c */
+    value unrooted_callback = ml_values->callback;
+    caml_modify_generational_global_root(&ml_values->callback, Val_unit);
+    value res_or_exn = caml_callback_exn(unrooted_callback, Val_unit);
+    value res = make_finished(res_or_exn);
+    sync_result(ml_values->term_sync, res);
+
+    sync_mutex mut = Mutex_val(*Term_mutex(ml_values->term_sync));
     domain_terminate();
 
-    /* This domain currently holds the [term_mutex], and has signaled all the
-       waiting domains to be woken up. We unlock the [term_mutex] to release
-       the joining domains. The unlock is done after [domain_terminate] to
-       ensure that this domain has released all of its runtime state. */
-    caml_mutex_unlock(Mutex_val(ml_values->term_mutex));
+    /* This domain currently holds [mut], and has signaled all the
+       waiting domains to be woken up. We unlock [mut] to release the
+       joining domains. The unlock is done after [domain_terminate] to
+       ensure that this domain has released all of its runtime state.
+       We call [caml_mutex_unlock] directly instead of
+       [caml_ml_mutex_unlock] because the domain no longer exists at
+       this point. */
+    caml_mutex_unlock(mut);
 
-    /* [ml_values] must be freed after unlocking [term_mutex]. This ensures
-       that [term_mutex] is only removed from the root set after [term_mutex]
-       is unlocked. Otherwise, there is a risk of [term_mutex] being destroyed
-       by [caml_mutex_finalize] finaliser while it remains locked, leading to
-       undefined behaviour. */
+    /* [ml_values] must be freed after unlocking [mut]. This ensures
+       that [term_sync] is only removed from the root set after [mut]
+       is unlocked. Otherwise, there is a risk of [mut] being
+       destroyed by [caml_mutex_finalize] finaliser while it remains
+       locked, leading to undefined behaviour. */
     free_domain_ml_values(ml_values);
   } else {
     caml_gc_log("Failed to create domain");
@@ -1126,45 +1243,26 @@ static void* domain_thread_func(void* v)
   return 0;
 }
 
-CAMLprim value caml_domain_spawn(value callback, value mutex)
+CAMLprim value caml_domain_spawn(value callback, value term_sync)
 {
-  CAMLparam2 (callback, mutex);
+  CAMLparam2 (callback, term_sync);
   struct domain_startup_params p;
   pthread_t th;
   int err;
-#ifndef _WIN32
-  sigset_t mask, old_mask;
-#endif
 
 #ifndef NATIVE_CODE
   if (caml_debugger_in_use)
     caml_fatal_error("ocamldebug does not support spawning multiple domains");
 #endif
-  p.parent = &domain_self->interruptor;
+  p.parent = domain_self;
   p.status = Dom_starting;
 
   p.ml_values =
-      (struct domain_ml_values*) caml_stat_alloc_noexc(
+      (struct domain_ml_values*) caml_stat_alloc(
                                     sizeof(struct domain_ml_values));
-  if (!p.ml_values) {
-    caml_failwith("failed to create ml values for domain thread");
-  }
-  init_domain_ml_values(p.ml_values, callback, mutex);
+  init_domain_ml_values(p.ml_values, callback, term_sync);
 
-/* We block all signals while we spawn the new domain. This is because
-   pthread_create inherits the current signals set, and we want to avoid a
-   signal handler being triggered in the new domain before the domain_state is
-   fully populated. */
-#ifndef _WIN32
-  sigfillset(&mask);
-  pthread_sigmask(SIG_BLOCK, &mask, &old_mask);
-  p.mask = &old_mask;
-#endif
   err = pthread_create(&th, 0, domain_thread_func, (void*)&p);
-#ifndef _WIN32
-  /* We can restore the signal mask we had initially now. */
-  pthread_sigmask(SIG_SETMASK, &old_mask, NULL);
-#endif
 
   if (err) {
     caml_failwith("failed to create domain thread");
@@ -1208,11 +1306,6 @@ CAMLprim value caml_ml_domain_id(value unit)
   return Val_long(domain_self->interruptor.unique_id);
 }
 
-CAMLprim value caml_ml_domain_unique_token (value unit)
-{
-  return Val_unit;
-}
-
 /* sense-reversing barrier */
 #define BARRIER_SENSE_BIT 0x100000
 
@@ -1520,26 +1613,69 @@ int caml_try_run_on_all_domains_async(
                                                  leader_setup, 0, 0);
 }
 
-void caml_interrupt_self(void) {
-  interrupt_domain(&domain_self->interruptor);
+void caml_interrupt_self(void)
+{
+  interrupt_domain_local(Caml_state);
 }
 
+/* async-signal-safe */
+void caml_interrupt_all_signal_safe(void)
+{
+  for (dom_internal *d = all_domains; d < &all_domains[Max_domains]; d++) {
+    /* [all_domains] is an array of values. So we can access
+       [interrupt_word] directly without synchronisation other than
+       with other people who access the same [interrupt_word].*/
+    atomic_uintnat * interrupt_word =
+      atomic_load_acquire(&d->interruptor.interrupt_word);
+    /* Early exit: if the current domain was never initialized, then
+       neither have been any of the remaining ones. */
+    if (interrupt_word == NULL) return;
+    interrupt_domain(&d->interruptor);
+  }
+}
+
+/* To avoid any risk of forgetting an action through a race,
+   [caml_reset_young_limit] is the only way (apart from setting
+   young_limit to -1 for immediate interruption) through which
+   [young_limit] can be modified. We take care here of possible
+   races. */
 void caml_reset_young_limit(caml_domain_state * dom_st)
 {
-  CAMLassert ((uintnat)dom_st->young_ptr >= (uintnat)dom_st->young_trigger);
+  /* An interrupt might have been queued in the meanwhile; the
+     atomic_exchange achieves the proper synchronisation with the
+     reads that follow (an atomic_store is not enough). */
+  value *trigger = dom_st->young_trigger > dom_st->memprof_young_trigger ?
+          dom_st->young_trigger : dom_st->memprof_young_trigger;
+  CAMLassert ((uintnat)dom_st->young_ptr >=
+              (uintnat)dom_st->memprof_young_trigger);
+  CAMLassert ((uintnat)dom_st->young_ptr >=
+              (uintnat)dom_st->young_trigger);
   /* An interrupt might have been queued in the meanwhile; this
      achieves the proper synchronisation. */
-  atomic_exchange(&dom_st->young_limit, (uintnat)dom_st->young_trigger);
+  atomic_exchange(&dom_st->young_limit, (uintnat)trigger);
+
+  /* For non-delayable asynchronous actions, we immediately interrupt
+     the domain again. */
   dom_internal * d = &all_domains[dom_st->id];
   if (atomic_load_relaxed(&d->interruptor.interrupt_pending)
       || dom_st->requested_minor_gc
       || dom_st->requested_major_slice
-      || dom_st->major_slice_epoch < atomic_load (&caml_major_slice_epoch)
-      || atomic_load_relaxed(&dom_st->requested_external_interrupt)
-      || dom_st->action_pending) {
-    atomic_store_release(&dom_st->young_limit, (uintnat)-1);
-    CAMLassert(caml_check_gc_interrupt(dom_st));
+      || dom_st->major_slice_epoch < atomic_load (&caml_major_slice_epoch)) {
+    interrupt_domain_local(dom_st);
   }
+  /* We might be here due to a recently-recorded signal or forced
+     systhread switching, so we need to remember that we must run
+     signal handlers or systhread's yield. In addition, in the case of
+     long-running C code (that may regularly poll with
+     caml_process_pending_actions), we want to force a query of all
+     callbacks at every minor collection or major slice (similarly to
+     the OCaml behaviour). */
+  caml_set_action_pending(dom_st);
+}
+
+void caml_update_young_limit_after_c_call(caml_domain_state * dom_st)
+{
+  if (CAMLunlikely(dom_st->action_pending)) interrupt_domain_local(dom_st);
 }
 
 Caml_inline void advance_global_major_slice_epoch (caml_domain_state* d)
@@ -1555,34 +1691,15 @@ Caml_inline void advance_global_major_slice_epoch (caml_domain_state* d)
   if (old_value != atomic_load (&caml_minor_collections_count)) {
     /* This domain is the first one to use up half of its minor heap arena
         in this minor cycle. Trigger major slice on other domains. */
-    if (caml_plat_try_lock(&all_domains_lock)) {
-      /* Note that this interrupt is best-effort. If we get the lock,
-         then interrupt all the domains. If not, either some other domain
-         is calling for a stop-the-world section interrupting all the
-         domains, or a domain is being created or terminated. All of these
-         actions also try to lock [all_domains_lock] mutex, and the above
-         lock acquisition may fail.
-
-         If we don't get the lock, we don't interrupt other domains. This is
-         acceptable since it does not affect safety but only liveness -- the
-         speed of the major gc. The other domains may themselves fill half of
-         their minor heap triggering a major slice, or will certainly do a
-         major slice right after their next minor GC when they observe that
-         their domain-local [Caml_state->major_slice_epoch] is less than the
-         global one [caml_major_slice_epoch]. */
-      for(int i = 0; i < stw_domains.participating_domains; i++) {
-        dom_internal * di = stw_domains.domains[i];
-        if (di->state != d) interrupt_domain(&di->interruptor);
-      }
-      caml_plat_unlock (&all_domains_lock);
-    }
+    caml_interrupt_all_signal_safe();
   }
 }
 
-static void global_major_slice_callback (caml_domain_state *domain,
-                                         void *unused,
-                                         int participating_count,
-                                         caml_domain_state **participating)
+static void stw_global_major_slice(
+  caml_domain_state *domain,
+  void *unused,
+  int participating_count,
+  caml_domain_state **participating)
 {
   domain->requested_major_slice = 1;
   /* Nothing else to do, as [stw_hander] will call [caml_poll_gc_work]
@@ -1637,16 +1754,13 @@ void caml_poll_gc_work(void)
 
   if (d->requested_global_major_slice) {
     if (caml_try_run_on_all_domains_async(
-          &global_major_slice_callback, NULL, NULL)){
+          &stw_global_major_slice, NULL, NULL)){
       d->requested_global_major_slice = 0;
     }
     /* If caml_try_run_on_all_domains_async fails, we'll try again next time
        caml_poll_gc_work is called. */
   }
 
-  if (atomic_load_acquire(&d->requested_external_interrupt)) {
-    caml_domain_external_interrupt_hook();
-  }
   caml_reset_young_limit(d);
 }
 
@@ -1664,6 +1778,14 @@ void caml_handle_gc_interrupt(void)
   caml_poll_gc_work();
 }
 
+/* Preemptive systhread switching */
+void caml_process_external_interrupt(void)
+{
+  if (atomic_load_acquire(&Caml_state->requested_external_interrupt)) {
+    caml_domain_external_interrupt_hook();
+  }
+}
+
 CAMLexport int caml_bt_is_in_blocking_section(void)
 {
   uintnat status = atomic_load_acquire(&domain_self->backup_thread_msg);
@@ -1720,48 +1842,29 @@ CAMLexport void caml_bt_exit_ocaml(void)
 }
 
 /* default handler for unix_fork, will be called by unix_fork. */
-static void caml_atfork_default(void) {
+static void caml_atfork_default(void)
+{
   caml_reset_domain_lock();
   caml_acquire_domain_lock();
+  /* FIXME: For best portability, the IO channel locks should be
+     reinitialised as well. (See comment in
+     caml_reset_domain_lock.) */
 }
 
 CAMLexport void (*caml_atfork_hook)(void) = caml_atfork_default;
 
-static void handover_ephemerons(caml_domain_state* domain_state)
-{
-  if (domain_state->ephe_info->todo == 0 &&
-      domain_state->ephe_info->live == 0 &&
-      domain_state->ephe_info->must_sweep_ephe == 0)
-    return;
-
-  caml_add_to_orphaned_ephe_list(domain_state->ephe_info);
-  CAMLassert (domain_state->ephe_info->live == 0);
-  CAMLassert (domain_state->ephe_info->todo == 0);
+static inline int domain_terminating(dom_internal *d) {
+  return d->interruptor.terminating;
 }
 
-static void handover_finalisers(caml_domain_state* domain_state)
+int caml_domain_terminating (caml_domain_state *dom_st)
 {
-  struct caml_final_info* f = domain_state->final_info;
-
-  if (f->todo_head != NULL || f->first.size != 0 || f->last.size != 0) {
-    /* have some final structures */
-    if (caml_gc_phase != Phase_sweep_and_mark_main) {
-      /* Force a major GC cycle to simplify constraints for
-       * handing over finalisers. */
-      caml_finish_major_cycle();
-      CAMLassert(caml_gc_phase == Phase_sweep_and_mark_main);
-    }
-    caml_add_orphaned_finalisers (f);
-    /* Create a dummy final info */
-    domain_state->final_info = caml_alloc_final_info();
-  }
-  caml_final_domain_terminate(domain_state);
+  return domain_terminating(&all_domains[dom_st->id]);
 }
 
 int caml_domain_is_terminating (void)
 {
-  struct interruptor* s = &domain_self->interruptor;
-  return s->terminating;
+  return domain_terminating(domain_self);
 }
 
 static void domain_terminate (void)
@@ -1769,19 +1872,10 @@ static void domain_terminate (void)
   caml_domain_state* domain_state = domain_self->state;
   struct interruptor* s = &domain_self->interruptor;
   int finished = 0;
-#ifndef _WIN32
-  sigset_t mask;
-#endif
 
   caml_gc_log("Domain terminating");
   s->terminating = 1;
 
-#ifndef _WIN32
-  /* Block all signals so that signal handlers do not run on this thread */
-  sigfillset(&mask);
-  pthread_sigmask(SIG_BLOCK, &mask, NULL);
-#endif
-
   /* Join ongoing systhreads, if necessary, and then run user-defined
      termination hooks. No OCaml code can run on this domain after
      this. */
@@ -1796,21 +1890,22 @@ static void domain_terminate (void)
        STW sections that has sent an interrupt to this domain. */
 
     caml_finish_marking();
-    handover_ephemerons(domain_state);
-    handover_finalisers(domain_state);
+
+    caml_orphan_ephemerons(domain_state);
+    caml_orphan_finalisers(domain_state);
 
     /* take the all_domains_lock to try and exit the STW participant set
        without racing with a STW section being triggered */
     caml_plat_lock(&all_domains_lock);
 
     /* The interaction of termination and major GC is quite subtle.
-     *
-     * At the end of the major GC, we decide the number of domains to mark and
-     * sweep for the next cycle. If a STW section has been started, it will
-     * require this domain to participate, which in turn could involve a
-     * major GC cycle. This would then require finish marking and sweeping
-     * again in order to decrement the globals [num_domains_to_mark] and
-     * [num_domains_to_sweep] (see major_gc.c).
+
+       At the end of the major GC, we decide the number of domains to mark and
+       sweep for the next cycle. If a STW section has been started, it will
+       require this domain to participate, which in turn could involve a major
+       GC cycle. This would then require finish marking and sweeping again in
+       order to decrement the globals [num_domains_to_mark] and
+       [num_domains_to_sweep] (see major_gc.c).
      */
 
     if (!caml_incoming_interrupts_queued() &&
@@ -1833,9 +1928,26 @@ static void domain_terminate (void)
 
       CAMLassert (domain_self->backup_thread_running);
       domain_self->backup_thread_running = 0;
+
+      /* We must signal domain termination before releasing [all_domains_lock]:
+         after that, this domain will no longer take part in STWs and emitting
+         an event could race with runtime events teardown. */
+      CAML_EV_LIFECYCLE(EV_DOMAIN_TERMINATE, getpid());
     }
     caml_plat_unlock(&all_domains_lock);
   }
+
+  /* domain_state may be re-used by a fresh domain here (now that we
+   * have done remove_from_stw_domains and released the
+   * all_domains_lock). However, domain_create() won't touch it until
+   * it has claimed the domain_lock, so we hang onto that while we are
+   * tearing down the state. */
+
+  /* Delete the domain state from statmemprof after any promotion
+   * (etc) done by this domain: any remaining memprof state will be
+   * handed over to surviving domains. */
+  caml_memprof_delete_domain(domain_state);
+
   /* We can not touch domain_self->interruptor after here
      because it may be reused */
   caml_remove_generational_global_root(&domain_state->dls_root);
@@ -1845,7 +1957,6 @@ static void domain_terminate (void)
   caml_free_intern_state();
   caml_free_extern_state();
   caml_teardown_major_gc();
-  CAML_EV_LIFECYCLE(EV_DOMAIN_TERMINATE, getpid());
 
   caml_teardown_shared_heap(domain_state->shared_heap);
   domain_state->shared_heap = 0;
@@ -1853,13 +1964,13 @@ static void domain_terminate (void)
   domain_state->minor_tables = 0;
 
   caml_orphan_alloc_stats(domain_state);
-  /* Heap stats were orphaned by caml_teardown_shared_heap above.
-     At this point, the stats of the domain must be empty;
-     we also clear the sampled copy.
+  /* Heap stats were orphaned by [caml_teardown_shared_heap] above.
+     At this point, the stats of the domain must be empty.
 
-     Note: We cannot call caml_collect_gc_stats_sample to clear the
-     sample at this point as the shared heap is gone. */
-  caml_clear_gc_stats_sample(domain_state);
+     The sampled copy was also cleared by the minor collection(s)
+     performed above at [caml_empty_minor_heaps_once()], see the
+     termination-specific logic in [caml_collect_gc_stats_sample_stw].
+  */
 
   /* TODO: can this ever be NULL? can we remove this check? */
   if(domain_state->current_stack != NULL) {
@@ -1902,6 +2013,18 @@ CAMLprim value caml_domain_dls_get(value unused)
   return Caml_state->dls_root;
 }
 
+CAMLprim value caml_domain_dls_compare_and_set(value old, value new)
+{
+  CAMLnoalloc;
+  value current = Caml_state->dls_root;
+  if (current == old) {
+    caml_modify_generational_global_root(&Caml_state->dls_root, new);
+    return Val_true;
+  } else {
+    return Val_false;
+  }
+}
+
 CAMLprim value caml_recommended_domain_count(value unused)
 {
   intnat n = -1;
