@@ -64,7 +64,11 @@ struct caml_intern_state {
 
   unsigned char * intern_input;
   /* Pointer to beginning of block holding input data,
-    if non-NULL this pointer will be freed by the cleanup function. */
+    if non-NULL this pointer will be freed by the cleanup function.
+
+    Allocated using malloc/free instead of stat_alloc/stat_free to
+    satisfy the expectations of caml_input_value_from_malloc users.
+  */
 
   asize_t obj_counter;
   /* Count how many objects seen so far */
@@ -88,8 +92,14 @@ struct caml_intern_state {
   /* 1 if the compressed format is in use, 0 otherwise */
 };
 
-/* Allocates the domain local intern state if needed */
-static struct caml_intern_state* get_intern_state (void)
+static void init_intern_stack(struct caml_intern_state* s)
+{
+  /* (Re)initialize the globals for next time around */
+  s->intern_stack = s->intern_stack_init;
+  s->intern_stack_limit = s->intern_stack + INTERN_STACK_INIT_SIZE;
+}
+
+static struct caml_intern_state* init_intern_state (void)
 {
   Caml_check_caml_state();
   struct caml_intern_state* s;
@@ -103,27 +113,38 @@ static struct caml_intern_state* get_intern_state (void)
   s->intern_input = NULL;
   s->obj_counter = 0;
   s->intern_obj_table = NULL;
-  s->intern_stack = s->intern_stack_init;
-  s->intern_stack_limit = s->intern_stack + INTERN_STACK_INIT_SIZE;
   s->intern_dest = NULL;
+  init_intern_stack(s);
 
   Caml_state->intern_state = s;
   return s;
 }
 
+static struct caml_intern_state* get_intern_state (void)
+{
+  Caml_check_caml_state();
+
+  if (Caml_state->intern_state == NULL)
+    caml_fatal_error (
+      "intern_state not initialized: it is likely that a caml_deserialize_* "
+      "function was called without going through caml_input_*."
+      );
+
+  return Caml_state->intern_state;
+}
+
 void caml_free_intern_state (void)
 {
-  if (Caml_state->intern_state != NULL)
+  if (Caml_state->intern_state != NULL) {
     caml_stat_free(Caml_state->intern_state);
-  Caml_state->intern_state = NULL;
+    Caml_state->intern_state = NULL;
+  }
 }
 
 static char * intern_resolve_code_pointer(unsigned char digest[16],
                                           asize_t offset);
 
-CAMLnoreturn_start
-static void intern_bad_code_pointer(unsigned char digest[16])
-CAMLnoreturn_end;
+CAMLnoret static void intern_bad_code_pointer(unsigned char digest[16]);
 
 Caml_inline unsigned char read8u(struct caml_intern_state* s)
 { return *s->intern_src++; }
@@ -219,16 +240,14 @@ static void intern_free_stack(struct caml_intern_state* s)
 {
   if (s->intern_stack != s->intern_stack_init) {
     caml_stat_free(s->intern_stack);
-    /* Reinitialize the globals for next time around */
-    s->intern_stack = s->intern_stack_init;
-    s->intern_stack_limit = s->intern_stack + INTERN_STACK_INIT_SIZE;
+    init_intern_stack(s);
   }
 }
 
 static void intern_cleanup(struct caml_intern_state* s)
 {
   if (s->intern_input != NULL) {
-     caml_stat_free(s->intern_input);
+     free(s->intern_input);
      s->intern_input = NULL;
   }
   if (s->intern_obj_table != NULL) {
@@ -240,6 +259,22 @@ static void intern_cleanup(struct caml_intern_state* s)
   intern_free_stack(s);
 }
 
+CAMLnoret static void intern_failwith2(const char * fun_name, const char * msg)
+{
+  caml_failwith_value(caml_alloc_sprintf("%s: %s", fun_name, msg));
+}
+
+CAMLnoret static void
+intern_cleanup_failwith3(struct caml_intern_state* s, const char * fun_name,
+                         const char * msg, const char * arg)
+{
+  /* Format the message before calling intern_cleanup, as arg may point
+     into intern buffers */
+  value v = caml_alloc_sprintf("%s: %s %s", fun_name, msg, arg);
+  intern_cleanup(s);
+  caml_failwith_value(v);
+}
+
 static void readfloat(struct caml_intern_state* s,
                       double * dest, unsigned int code)
 {
@@ -258,7 +293,7 @@ static void readfloat(struct caml_intern_state* s,
 #else
   /* Host is neither big nor little; permute as appropriate */
   if (code == CODE_DOUBLE_LITTLE)
-    Permute_64(dest, ARCH_FLOAT_ENDIANNESS, dest, 0x01234567)
+    Permute_64(dest, ARCH_FLOAT_ENDIANNESS, dest, 0x01234567);
   else
     Permute_64(dest, ARCH_FLOAT_ENDIANNESS, dest, 0x76543210);
 #endif
@@ -300,11 +335,7 @@ static void readfloats(struct caml_intern_state* s,
 #endif
 }
 
-CAMLnoreturn_start
-static void intern_stack_overflow(struct caml_intern_state*)
-CAMLnoreturn_end;
-
-static void intern_stack_overflow(struct caml_intern_state* s)
+CAMLnoret static void intern_stack_overflow(struct caml_intern_state* s)
 {
   caml_gc_message (0x04, "Stack overflow in un-marshaling value\n");
   intern_cleanup(s);
@@ -398,8 +429,7 @@ static value intern_alloc_obj(struct caml_intern_state* s, caml_domain_state* d,
     s->intern_dest += 1 + wosize;
   } else {
     p = caml_shared_try_alloc(d->shared_heap, wosize, tag,
-                              0, /* no reserved bits */
-                              0 /* not pinned */);
+                              0 /* no reserved bits */);
     d->allocated_words += Whsize_wosize(wosize);
     if (p == NULL) {
       intern_cleanup (s);
@@ -411,6 +441,7 @@ static value intern_alloc_obj(struct caml_intern_state* s, caml_domain_state* d,
 }
 
 static void intern_rec(struct caml_intern_state* s,
+                       const char * fun_name,
                        volatile value *dest)
 {
   unsigned int code;
@@ -518,7 +549,7 @@ static void intern_rec(struct caml_intern_state* s,
         break;
 #else
         intern_cleanup(s);
-        caml_failwith("input_value: integer too large");
+        intern_failwith2(fun_name, "integer too large");
         break;
 #endif
       case CODE_SHARED8:
@@ -619,20 +650,21 @@ static void intern_rec(struct caml_intern_state* s,
         continue;  /* with next iteration of main loop, skipping *dest = v */
       case OLD_CODE_CUSTOM:
         intern_cleanup(s);
-        caml_failwith("input_value: custom blocks serialized with "
-                      "OCaml 4.08.0 (or prior) are no longer supported");
+        intern_failwith2(fun_name, "custom blocks serialized with "
+                         "OCaml 4.08.0 (or prior) are no longer supported");
         break;
       case CODE_CUSTOM_LEN:
       case CODE_CUSTOM_FIXED: {
         uintnat expected_size, temp_size;
-        ops = caml_find_custom_operations((char *) s->intern_src);
+        const char * name = (const char *) s->intern_src;
+        ops = caml_find_custom_operations(name);
         if (ops == NULL) {
-          intern_cleanup(s);
-          caml_failwith("input_value: unknown custom block identifier");
+          intern_cleanup_failwith3
+            (s, fun_name, "unknown custom block identifier", name);
         }
         if (code == CODE_CUSTOM_FIXED && ops->fixed_length == NULL) {
-          intern_cleanup(s);
-          caml_failwith("input_value: expected a fixed-size custom block");
+          intern_cleanup_failwith3
+            (s, fun_name, "expected a fixed-size custom block", name);
         }
         while (*s->intern_src++ != 0) /*nothing*/;  /*skip identifier*/
 #ifdef ARCH_SIXTYFOUR
@@ -655,9 +687,8 @@ static void intern_rec(struct caml_intern_state* s,
         Custom_ops_val(v) = ops;
         size = ops->deserialize(Data_custom_val(v));
         if (size != expected_size) {
-          intern_cleanup(s);
-          caml_failwith(
-            "input_value: incorrect length of serialized custom block");
+          intern_cleanup_failwith3
+            (s, fun_name, "error while deserializing custom block", name);
         }
         if (s->intern_obj_table != NULL)
           s->intern_obj_table[s->obj_counter++] = v;
@@ -669,7 +700,7 @@ static void intern_rec(struct caml_intern_state* s,
       }
       default:
         intern_cleanup(s);
-        caml_failwith("input_value: ill-formed message");
+        intern_failwith2(fun_name, "ill-formed message");
       }
     }
   }
@@ -719,14 +750,6 @@ struct marshal_header {
   int compressed;
 };
 
-static void intern_failwith2(const char * fun_name, const char * msg)
-{
-  char errmsg[100];
-  errmsg[sizeof(errmsg) - 1] = 0;
-  snprintf(errmsg, sizeof(errmsg) - 1, "%s: %s", fun_name, msg);
-  caml_failwith(errmsg);
-}
-
 static void caml_parse_header(struct caml_intern_state* s,
                               const char * fun_name,
                               /*out*/ struct marshal_header * h)
@@ -834,7 +857,7 @@ value caml_input_val(struct channel *chan)
   struct marshal_header h;
   char * block;
   value res;
-  struct caml_intern_state* s = get_intern_state ();
+  struct caml_intern_state* s = init_intern_state ();
 
   if (! caml_channel_binary_mode(chan))
     caml_failwith("input_value: not a binary channel");
@@ -866,9 +889,12 @@ value caml_input_val(struct channel *chan)
      can take place (via context switching in systhreads),
      and the context [s] may change.  So, wait until all I/O is over
      before using the context [s] again. */
-  block = caml_stat_alloc(h.data_len);
+  block = malloc(h.data_len);
+  if (block == NULL) {
+    caml_raise_out_of_memory();
+  }
   if (caml_really_getblock(chan, block, h.data_len) < h.data_len) {
-    caml_stat_free(block);
+    free(block);
     caml_failwith("input_value: truncated object");
   }
   /* Initialize global state */
@@ -876,7 +902,7 @@ value caml_input_val(struct channel *chan)
   intern_decompress_input(s, "input_value", &h);
   intern_alloc_storage(s, h.whsize, h.num_objects);
   /* Fill it in */
-  intern_rec(s, &res);
+  intern_rec(s, "input_value", &res);
   return intern_end(s, res);
 }
 
@@ -886,9 +912,9 @@ CAMLprim value caml_input_value(value vchan)
   struct channel * chan = Channel(vchan);
   CAMLlocal1 (res);
 
-  Lock(chan);
+  caml_channel_lock(chan);
   res = caml_input_val(chan);
-  Unlock(chan);
+  caml_channel_unlock(chan);
   CAMLreturn (res);
 }
 
@@ -905,7 +931,7 @@ CAMLexport value caml_input_val_from_bytes(value str, intnat ofs)
   CAMLparam1 (str);
   CAMLlocal1 (obj);
   struct marshal_header h;
-  struct caml_intern_state* s = get_intern_state ();
+  struct caml_intern_state* s = init_intern_state ();
 
   /* Initialize global state */
   intern_init(s, &Byte_u(str, ofs), NULL);
@@ -918,7 +944,7 @@ CAMLexport value caml_input_val_from_bytes(value str, intnat ofs)
   /* Decompress if needed */
   intern_decompress_input(s, "input_val_from_string", &h);
   /* Fill it in */
-  intern_rec(s, &obj);
+  intern_rec(s, "input_val_from_string", &obj);
   CAMLreturn (intern_end(s, obj));
 }
 
@@ -936,14 +962,14 @@ static value input_val_from_block(struct caml_intern_state* s,
   /* Allocate result */
   intern_alloc_storage(s, h->whsize, h->num_objects);
   /* Fill it in */
-  intern_rec(s, &obj);
+  intern_rec(s, "input_val_from_block", &obj);
   return (intern_end(s, obj));
 }
 
 CAMLexport value caml_input_value_from_malloc(char * data, intnat ofs)
 {
   struct marshal_header h;
-  struct caml_intern_state* s = get_intern_state ();
+  struct caml_intern_state* s = init_intern_state ();
 
   intern_init(s, data + ofs, data);
   caml_parse_header(s, "input_value_from_malloc", &h);
@@ -954,7 +980,7 @@ CAMLexport value caml_input_value_from_malloc(char * data, intnat ofs)
 CAMLexport value caml_input_value_from_block(const char * data, intnat len)
 {
   struct marshal_header h;
-  struct caml_intern_state* s = get_intern_state ();
+  struct caml_intern_state* s = init_intern_state ();
 
   /* Initialize global state */
   intern_init(s, data, NULL);
@@ -984,7 +1010,7 @@ CAMLprim value caml_marshal_data_size(value buff, value ofs)
   uint32_t magic;
   int header_len;
   uintnat data_len;
-  struct caml_intern_state *s = get_intern_state ();
+  struct caml_intern_state *s = init_intern_state ();
 
   s->intern_src = &Byte_u(buff, Long_val(ofs));
   magic = read32u(s);
