@@ -151,7 +151,7 @@ static value alloc_shared(caml_domain_state* d,
                           mlsize_t wosize, tag_t tag, reserved_t reserved)
 {
   void* mem = caml_shared_try_alloc(d->shared_heap, wosize, tag,
-                                    reserved, 0 /* not pinned */);
+                                    reserved);
   d->allocated_words += Whsize_wosize(wosize);
   if (mem == NULL) {
     caml_fatal_error("allocation failure during minor GC");
@@ -171,6 +171,7 @@ static void spin_on_header(value v) {
   }
 }
 
+CAMLno_tsan_for_perf
 Caml_inline header_t get_header_val(value v) {
   header_t hd = atomic_load_acquire(Hp_atomic_val(v));
   if (!Is_update_in_progress(hd))
@@ -270,11 +271,12 @@ static void oldify_one (void* st_v, value v, volatile value *p)
 
   if (tag == Cont_tag) {
     value stack_value = Field(v, 0);
-    CAMLassert(Wosize_hd(hd) == 1 && infix_offset == 0);
-    result = alloc_shared(st->domain, 1, Cont_tag, Reserved_hd(hd));
+    CAMLassert(Wosize_hd(hd) == 2 && infix_offset == 0);
+    result = alloc_shared(st->domain, 2, Cont_tag, Reserved_hd(hd));
     if( try_update_object_header(v, p, result, 0) ) {
       struct stack_info* stk = Ptr_val(stack_value);
-      Field(result, 0) = Val_ptr(stk);
+      Field(result, 0) = stack_value;
+      Field(result, 1) = Field(v, 1);
       if (stk != NULL) {
         caml_scan_stack(&oldify_one, oldify_scanning_flags, st,
                         stk, 0);
@@ -287,6 +289,7 @@ static void oldify_one (void* st_v, value v, volatile value *p)
                                     caml_global_heap_state.MARKED);
       #ifdef DEBUG
       Field(result, 0) = Val_long(1);
+      Field(result, 1) = Val_long(1);
       #endif
     }
   } else if (tag < Infix_tag) {
@@ -379,6 +382,7 @@ static void oldify_one (void* st_v, value v, volatile value *p)
    Note that [oldify_one] itself is called by oldify_mopup, so we
    have to be careful to remove the first entry from the list before
    oldifying its fields. */
+CAMLno_tsan_for_perf
 static void oldify_mopup (struct oldify_state* st, int do_ephemerons)
 {
   value v, new_v, f;
@@ -542,7 +546,11 @@ void caml_empty_minor_heap_promote(caml_domain_state* domain,
 
       for( r = ref_start ; r < foreign_major_ref->ptr && r < ref_end ; r++ )
       {
-        oldify_one (&st, **r, *r);
+        /* Because the work on the remembered set is shared, other threads may
+           attempt to promote the same value; this is fine, but we need the
+           writes and reads (here, `*pr`) to be at least `volatile`. */
+        value_ptr pr = *r;
+        oldify_one (&st, *pr, pr);
         remembered_roots++;
       }
 
@@ -621,10 +629,9 @@ void caml_empty_minor_heap_promote(caml_domain_state* domain,
   domain->stat_minor_words += Wsize_bsize (minor_allocated_bytes);
   domain->stat_promoted_words += domain->allocated_words - prev_alloc_words;
 
-  /* gc stats may be accessed unsynchronised by mutator code, so we collect the
-     sample before arriving at the barrier, which ensures that it doesn't race
-  */
-  caml_collect_gc_stats_sample(domain);
+  /* Must be called during the STW section -- before any mutators
+     start running, so before arriving at the barrier. */
+  caml_collect_gc_stats_sample_stw(domain);
 
   /* The code above is synchronised with other domains by the barrier below,
      which is split into two steps, "arriving" and "leaving". When the final
@@ -751,6 +758,7 @@ caml_stw_empty_minor_heap_no_major_slice(caml_domain_state* domain,
   CAML_EV_BEGIN(EV_MINOR_CLEAR);
   caml_gc_log("running stw empty_minor_heap_domain_clear");
   caml_empty_minor_heap_domain_clear(domain);
+
 #ifdef DEBUG
   {
     for (uintnat* p = initial_young_ptr; p < (uintnat*)domain->young_end; ++p)
@@ -771,10 +779,11 @@ static void caml_stw_empty_minor_heap (caml_domain_state* domain, void* unused,
 }
 
 /* must be called within a STW section  */
-void caml_empty_minor_heap_no_major_slice_from_stw(caml_domain_state* domain,
-                                            void* unused,
-                                             int participating_count,
-                                             caml_domain_state** participating)
+void caml_empty_minor_heap_no_major_slice_from_stw(
+  caml_domain_state* domain,
+  void* unused,
+  int participating_count,
+  caml_domain_state** participating)
 {
   barrier_status b = caml_global_barrier_begin();
   if( caml_global_barrier_is_final(b) ) {
@@ -789,7 +798,7 @@ void caml_empty_minor_heap_no_major_slice_from_stw(caml_domain_state* domain,
 }
 
 /* must be called outside a STW section */
-int caml_try_stw_empty_minor_heap_on_all_domains (void)
+int caml_try_empty_minor_heap_on_all_domains (void)
 {
   #ifdef DEBUG
   CAMLassert(!caml_domain_is_in_stw());
@@ -817,7 +826,7 @@ void caml_empty_minor_heaps_once (void)
   /* To handle the case where multiple domains try to execute a minor gc
      STW section */
   do {
-    caml_try_stw_empty_minor_heap_on_all_domains();
+    caml_try_empty_minor_heap_on_all_domains();
   } while (saved_minor_cycle == atomic_load(&caml_minor_cycles_started));
 }
 
@@ -841,11 +850,9 @@ void caml_alloc_small_dispatch (caml_domain_state * dom_st,
          asynchronous callbacks. */
       caml_raise_if_exception(caml_do_pending_actions_exn());
     else {
+      /* In the case of allocations performed from C, only perform
+         non-delayable actions. */
       caml_handle_gc_interrupt();
-      /* In the case of long-running C code that regularly polls with
-         [caml_process_pending_actions], still force a query of all
-         callbacks at every minor collection or major slice. */
-      dom_st->action_pending = 1;
     }
 
     /* Now, there might be enough room in the minor heap to do our
