@@ -1,5 +1,4 @@
-(* TEST
-*)
+(* TEST *)
 
 open Domain
 
@@ -30,9 +29,9 @@ let test_parallel_spawn () =
   done
 
 let () =
-  let running = ref true in
+  let running = Atomic.make true in
   let rec run_until_stop fn () =
-    while !running do
+    while Atomic.get running do
       fn ();
     done
   in
@@ -42,7 +41,7 @@ let () =
 
   test_parallel_spawn ();
 
-  running := false;
+  Atomic.set running false;
   join domain_minor_gc;
   join domain_major_gc;
 
