@@ -1,5 +1,5 @@
 (* TEST
-  * expect
+ expect;
 *)
 
 
@@ -41,7 +41,7 @@ module M = F(X)(Z)
 Line 1, characters 11-18:
 1 | module M = F(X)(Z)
                ^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          X Z
        do not match these parameters:
@@ -100,7 +100,7 @@ Error: Signature mismatch:
          sig module F : functor (X : sig type t end) -> sig end end
        is not included in
          sig module F : functor (X : sig end) -> sig end end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S1) -> ...
        is not included in
@@ -109,7 +109,7 @@ Error: Signature mismatch:
          $S1 = sig type t end
        does not include
          sig end
-       The type `t' is required but not provided
+       The type "t" is required but not provided
 |}]
 
 module F(X:sig type t end) = struct end
@@ -120,7 +120,7 @@ Line 2, characters 11-31:
 2 | module M = F(struct type x end)
                ^^^^^^^^^^^^^^^^^^^^
 Error: Modules do not match: sig type x end is not included in sig type t end
-     The type `t' is required but not provided
+     The type "t" is required but not provided
 |}]
 
 module F(X:sig type x end)(Y:sig type y end)(Z:sig type z end) = struct
@@ -134,7 +134,7 @@ module F :
 Line 4, characters 9-18:
 4 | type u = F(X)(Z).t
              ^^^^^^^^^
-Error: The functor application F(X)(Z) is ill-typed.
+Error: The functor application "F(X)(Z)" is ill-typed.
        These arguments:
          X Z
        do not match these parameters:
@@ -152,7 +152,7 @@ module F : functor () (X : sig type t end) -> sig end
 Line 2, characters 11-16:
 2 | module M = F()()
                ^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          () ()
        do not match these parameters:
@@ -179,7 +179,7 @@ Error: Signature mismatch:
            module F :
              functor (X : sig type x end) (X : sig type y end) -> sig end
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S2) -> ...
        is not included in
@@ -204,7 +204,7 @@ end
 Line 9, characters 13-20:
 9 |   module M = F(Y)(X)
                  ^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          Ctx.Y Ctx.X
        do not match these parameters:
@@ -221,7 +221,7 @@ module Ord : sig type t = unit val compare : 'a -> 'b -> int end
 Line 2, characters 11-29:
 2 | module M = Map.Make(Ord)(Ord)
                ^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "Map.Make" is ill-typed.
        These arguments:
          Ord Ord
        do not match these parameters:
@@ -250,7 +250,7 @@ module K : sig type x = X.x type y = Y.y end
 Line 10, characters 11-73:
 10 | module M = F(K)(struct type x = K.x end)( (* struct type z = K.y end *) )
                 ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          K $S2 ()
        do not match these parameters:
@@ -265,7 +265,7 @@ module M = F(K)(struct type y = K.y end)
 Line 1, characters 11-40:
 1 | module M = F(K)(struct type y = K.y end)
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          K $S3
        do not match these parameters:
@@ -288,7 +288,7 @@ Lines 2-5, characters 2-30:
 3 |     (struct include X include Y end)
 4 |     (struct type x = K.x end)
 5 |     (struct type yy = K.y end)
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          $S1 $S2 $S3
        do not match these parameters:
@@ -299,7 +299,7 @@ Error: The functor application is ill-typed.
             $S3 : sig type yy = K.y end
           is not included in
             $T3 = sig type y = A.y end
-          The type `y' is required but not provided
+          The type "y" is required but not provided
 |}]
 
 
@@ -323,7 +323,7 @@ module Defs :
 Line 13, characters 19-33:
 13 | module Missing_X = F(M.N)(Defs.Y)
                         ^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          M.N Defs.Y
        do not match these parameters:
@@ -339,7 +339,7 @@ module Too_many_Xs = F(M.N)(Defs.X)(Defs.X)(Defs.Y)
 Line 1, characters 21-51:
 1 | module Too_many_Xs = F(M.N)(Defs.X)(Defs.X)(Defs.Y)
                          ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          M.N Defs.X Defs.X Defs.Y
        do not match these parameters:
@@ -361,7 +361,7 @@ module Y : sig type y = float end
 Line 3, characters 23-67:
 3 | module Missing_X_bis = F(struct type x = int type y = float end)(Y)
                            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          $S1 Y
        do not match these parameters:
@@ -377,7 +377,7 @@ module Too_many_Xs_bis = F(struct type x = int type y = float end)(X)(X)(Y)
 Line 1, characters 25-75:
 1 | module Too_many_Xs_bis = F(struct type x = int type y = float end)(X)(X)(Y)
                              ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          $S1 X X Y
        do not match these parameters:
@@ -459,13 +459,13 @@ Error: Signature mismatch:
        does not match
          module type S = sig type t end
        The second module type is not included in the first
-       At position module type S = <here>
+       At position "module type S = <here>"
        Module types do not match:
          sig type t end
        is not equal to
          sig type s type t end
-       At position module type S = <here>
-       The type `s' is required but not provided
+       At position "module type S = <here>"
+       The type "s" is required but not provided
 |}]
 
 module M: sig
@@ -488,13 +488,13 @@ Error: Signature mismatch:
        does not match
          module type S = sig type t type u end
        The first module type is not included in the second
-       At position module type S = <here>
+       At position "module type S = <here>"
        Module types do not match:
          sig type t end
        is not equal to
          sig type t type u end
-       At position module type S = <here>
-       The type `u' is required but not provided
+       At position "module type S = <here>"
+       The type "u" is required but not provided
 |}]
 
 
@@ -513,7 +513,7 @@ module F : functor (X : x) (B : b) (Y : y) -> sig type t end
 Line 8, characters 15-57:
 8 |     module U = F(struct type x end)(B)(struct type w end)
                    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          $S1 P.B $S3
        do not match these parameters:
@@ -524,9 +524,9 @@ Error: The functor application is ill-typed.
           is not included in
             b/2
           Line 5, characters 2-15:
-            Definition of module type b
+            Definition of module type "b"
           Line 2, characters 0-13:
-            Definition of module type b/2
+            Definition of module type "b/2"
        3. Modules do not match: $S3 : sig type w end is not included in y
 |}]
 
@@ -545,9 +545,9 @@ Line 6, characters 13-19:
                  ^^^^^^
 Error: Modules do not match: a is not included in a/2
      Line 3, characters 2-15:
-       Definition of module type a
+       Definition of module type "a"
      Line 1, characters 0-13:
-       Definition of module type a/2
+       Definition of module type "a/2"
 |}]
 
 
@@ -574,7 +574,7 @@ Error: Signature mismatch:
          end
        is not included in
          sig module F : functor (X : a) (Y : a) -> sig end end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : aa) (Y : a) -> ...
        is not included in
@@ -585,9 +585,9 @@ Error: Signature mismatch:
           does not include
             a/2
           Line 4, characters 2-15:
-            Definition of module type a
+            Definition of module type "a"
           Line 1, characters 0-13:
-            Definition of module type a/2
+            Definition of module type "a/2"
 |}]
 
 module X: functor ( X: sig end) -> sig end = functor(X: Set.OrderedType) -> struct end
@@ -604,9 +604,9 @@ Error: Signature mismatch:
          Set.OrderedType
        does not include
          sig end
-       The type `t' is required but not provided
+       The type "t" is required but not provided
        File "set.mli", line 52, characters 4-10: Expected declaration
-       The value `compare' is required but not provided
+       The value "compare" is required but not provided
        File "set.mli", line 55, characters 4-31: Expected declaration
 |}]
 
@@ -674,7 +674,7 @@ Error: Signature mismatch:
                       sig end)
                -> sig end
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S1) (Y : $S2) (Z : $S3) -> ...
        is not included in
@@ -695,7 +695,7 @@ Error: Signature mismatch:
                $S2 = sig type yb end
              does not include
                $T2 = sig type ybb end
-             The type `yb' is required but not provided
+             The type "yb" is required but not provided
        3. Module types do not match:
             $S3 =
             functor (A : sig type za end) (B : sig type zbb end) -> sig end
@@ -765,7 +765,7 @@ Error: Signature mismatch:
                       sig end)
                -> sig end
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S1) (Y : $S2) -> ...
        is not included in
@@ -838,7 +838,7 @@ Error: Signature mismatch:
                       sig end)
                -> sig end
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S1) (Y : $S2) (Z : $S3) -> ...
        is not included in
@@ -857,7 +857,7 @@ Error: Signature mismatch:
                $S1 = sig type xa end
              does not include
                $T1 = sig type xaa end
-             The type `xa' is required but not provided
+             The type "xa" is required but not provided
           2. Module types $S2 and $T2 match
        2. Module types do not match:
             $S2 =
@@ -954,7 +954,7 @@ Error: Signature mismatch:
                  end
              end
          end
-       In module B:
+       In module "B":
        Modules do not match:
          sig module C = B.C end
        is not included in
@@ -972,7 +972,7 @@ Error: Signature mismatch:
                  end
              end
          end
-       In module B.C:
+       In module "B.C":
        Modules do not match:
          sig module D = B.C.D end
        is not included in
@@ -987,7 +987,7 @@ Error: Signature mismatch:
                  end
              end
          end
-       In module B.C.D:
+       In module "B.C.D":
        Modules do not match:
          sig module E = B.C.D.E end
        is not included in
@@ -999,7 +999,7 @@ Error: Signature mismatch:
                    sig type w end -> sig end
              end
          end
-       In module B.C.D.E:
+       In module "B.C.D.E":
        Modules do not match:
          sig module F = B.C.D.E.F end
        is not included in
@@ -1008,7 +1008,7 @@ Error: Signature mismatch:
              sig type x end -> sig type y end -> sig type z end ->
                sig type w end -> sig end
          end
-       In module B.C.D.E.F:
+       In module "B.C.D.E.F":
        Modules do not match:
          functor (X : $S1) (Y : $S3) (W : $S4) -> ...
        is not included in
@@ -1050,7 +1050,7 @@ module type Arg =
 Line 14, characters 11-29:
 14 |   type u = G(X)(Y)(X)(Y)(X).t
                 ^^^^^^^^^^^^^^^^^^
-Error: The functor application G(X)(Y)(X)(Y)(X) is ill-typed.
+Error: The functor application "G(X)(Y)(X)(Y)(X)" is ill-typed.
        These arguments:
          A.X A.Y A.X A.Y A.X
        do not match these parameters:
@@ -1157,7 +1157,7 @@ Error: Signature mismatch:
               type meet
               type again
             end
-          The type `tree' is required but not provided
+          The type "tree" is required but not provided
        2. Module types do not match:
             $S2 =
             sig type in_ val thunder : in_ val lightning : in_ type pain end
@@ -1253,7 +1253,7 @@ module W = F(PF)(PF)(PF)(PF)(PF)(F)
 Line 1, characters 11-35:
 1 | module W = F(PF)(PF)(PF)(PF)(PF)(F)
                ^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          PF PF PF PF PF F
        do not match these parameters:
@@ -1319,7 +1319,7 @@ module Choose_one = F(Add_one')(Add_three)(A)(A)(A)
 Line 1, characters 20-51:
 1 | module Choose_one = F(Add_one')(Add_three)(A)(A)(A)
                         ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          Add_one' Add_three A A A
        do not match these parameters:
@@ -1333,7 +1333,7 @@ Error: The functor application is ill-typed.
        5. Module A matches the expected module type arg
 |}]
 
-(** Known lmitation: we choose the wrong environment without the
+(** Known limitation: we choose the wrong environment without the
     error on Add_one
 **)
 module Mislead_chosen_one = F(Add_one)(Add_three)(A)(A)(A)
@@ -1341,7 +1341,7 @@ module Mislead_chosen_one = F(Add_one)(Add_three)(A)(A)(A)
 Line 1, characters 28-58:
 1 | module Mislead_chosen_one = F(Add_one)(Add_three)(A)(A)(A)
                                 ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          Add_one Add_three A A A
        do not match these parameters:
@@ -1403,7 +1403,7 @@ Error: Signature mismatch:
                     end)
                -> X.t
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S1) (Z : $S3) -> ...
        is not included in
@@ -1456,7 +1456,7 @@ Error: Signature mismatch:
                     end)
                -> X.T -> X.T -> X.T
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (Wrong : $S1) (X : $S2) X.T X.T -> ...
        is not included in
@@ -1526,7 +1526,7 @@ Error: Signature mismatch:
                       end)
                -> X.F(X.Y)(X.Y.M).res
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (Arg : $S1) (X : $S2) (Res : X.T) (Res : X.T) (Res :
          X.T) -> ...
@@ -1536,7 +1536,7 @@ Error: Signature mismatch:
             $S1 = sig type wrong end
           does not include
             sig end
-          The type `wrong' is required but not provided
+          The type "wrong" is required but not provided
        2. Module types $S2 and $T2 match
        3. An extra argument is provided of module type X.T
        4. Module types X.T and X.T match
@@ -1563,7 +1563,7 @@ module Z : sig type t = Z of int end
 Line 9, characters 13-48:
 9 | module Error=F(X)(struct type t = int end)(Y)(Z)
                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          X ... Y Z
        do not match these parameters:
@@ -1579,10 +1579,10 @@ Error: The functor application is ill-typed.
           is not included in
             type t = Y of X.t
           Constructors do not match:
-            Y of int
+            "Y of int"
           is not the same as:
-            Y of X.t
-          The type int is not equal to the type X.t
+            "Y of X.t"
+          The type "int" is not equal to the type "X.t"
        4. Modules do not match:
             Z : sig type t = Z.t = Z of int end
           is not included in
@@ -1592,10 +1592,10 @@ Error: The functor application is ill-typed.
           is not included in
             type t = Z of X.t
           Constructors do not match:
-            Z of int
+            "Z of int"
           is not the same as:
-            Z of X.t
-          The type int is not equal to the type X.t
+            "Z of X.t"
+          The type "int" is not equal to the type "X.t"
 |}]
 
 (** Final state in the presence of extensions
@@ -1643,7 +1643,7 @@ type broken1 = Bar(B)(FiveArgsExt)(B)(AExt).a
 Line 1, characters 15-45:
 1 | type broken1 = Bar(B)(FiveArgsExt)(B)(AExt).a
                    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application Bar(B)(FiveArgsExt)(B)(AExt) is ill-typed.
+Error: The functor application "Bar(B)(FiveArgsExt)(B)(AExt)" is ill-typed.
        These arguments:
          B FiveArgsExt B AExt
        do not match these parameters:
@@ -1652,7 +1652,7 @@ Error: The functor application Bar(B)(FiveArgsExt)(B)(AExt) is ill-typed.
             B : sig type b = B.b end
           is not included in
             A
-          The type `a' is required but not provided
+          The type "a" is required but not provided
        2. Module FiveArgsExt matches the expected module type Ext
        3. Module B matches the expected module type B
        4. Module AExt matches the expected module type Ext
@@ -1663,7 +1663,7 @@ type broken2 = Bar(A)(FiveArgsExt)(TY)(TY)(TY)(TY)(TY).a
 Line 1, characters 15-56:
 1 | type broken2 = Bar(A)(FiveArgsExt)(TY)(TY)(TY)(TY)(TY).a
                    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application Bar(A)(FiveArgsExt)(TY)(TY)(TY)(TY)(TY) is ill-typed.
+Error: The functor application "Bar(A)(FiveArgsExt)(TY)(TY)(TY)(TY)(TY)" is ill-typed.
        These arguments:
          A FiveArgsExt TY TY TY TY TY
        do not match these parameters:
@@ -1734,7 +1734,7 @@ Line 2, characters 11-40:
 2 | module R = F(struct end[@warning "-73"]);;
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 Error: Modules do not match: sig end is not included in A
-     The type `a' is required but not provided
+     The type "a" is required but not provided
 |}]
 
 module F()(X:empty)()(Y:A) = struct end
@@ -1745,7 +1745,7 @@ module F : functor () (X : empty) () (Y : A) -> sig end
 Line 3, characters 2-73:
 3 |   F(struct end[@warning "-73"])(struct end)(struct end[@warning "-73"])();;
       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          (struct end) (struct end) (struct end) ()
        do not match these parameters:
@@ -1765,7 +1765,7 @@ module F : functor (X : empty) -> sig end
 Line 3, characters 2-17:
 3 |   F(struct end)();;
       ^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          (struct end) ()
        do not match these parameters:
@@ -1808,7 +1808,7 @@ Lines 15-21, characters 2-8:
 19 |     (struct
 20 |       let f x = x   (* this is bogus! *)
 21 |     end)
-Error: The functor application is ill-typed.
+Error: This application of the functor "F" is ill-typed.
        These arguments:
          $S1 $S2
        do not match these parameters:
@@ -1844,7 +1844,7 @@ module G :
 Line 8, characters 11-52:
 8 | module R = G(struct end)(struct let f (x,_) = x end)
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-Error: The functor application is ill-typed.
+Error: This application of the functor "G" is ill-typed.
        These arguments:
          (struct end) $S2
        do not match these parameters:
@@ -1853,7 +1853,7 @@ Error: The functor application is ill-typed.
             (struct end)
           is not included in
             $T1 = sig type 'a t = 'a * 'a end
-          The type `t' is required but not provided
+          The type "t" is required but not provided
        2. Module $S2 matches the expected module type
 |}]
 
@@ -1880,7 +1880,7 @@ Lines 5-11, characters 11-6:
  9 |     ()
 10 |     ()
 11 |     ()
-Error: The functor application is ill-typed.
+Error: This application of the functor "With_expansion" is ill-typed.
        These arguments:
          $S1 () () ()
        do not match these parameters:
@@ -1906,7 +1906,7 @@ Lines 1-6, characters 12-6:
 4 |   end)
 5 |     ()
 6 |     ()
-Error: The functor application is ill-typed.
+Error: This application of the functor "With_expansion" is ill-typed.
        These arguments:
          $S1 () ()
        do not match these parameters:
@@ -1957,7 +1957,7 @@ Lines 18-25, characters 2-8:
 23 |     (struct
 24 |       let f x = x   (* this is fine *)
 25 |     end)
-Error: The functor application is ill-typed.
+Error: This application of the functor "H" is ill-typed.
        These arguments:
          $S1 $S2
        do not match these parameters:
@@ -1976,9 +1976,9 @@ Error: The functor application is ill-typed.
             val f : 'a -> 'a
           is not included in
             val f : 'a X.s -> 'a
-          The type 'a X.s -> 'a X.s is not compatible with the type
-            'a X.s -> 'a
-          Type 'a X.s is not compatible with type 'a
+          The type "'a X.s -> 'a X.s" is not compatible with the type
+            "'a X.s -> 'a"
+          Type "'a X.s" is not compatible with type "'a"
 |}]
 
 
@@ -2008,7 +2008,7 @@ Error: Signature mismatch:
              functor (X : sig type 'a t = 'a * 'a end)
                (Y : sig type 'a t = 'a X.t * 'a list end) -> sig end
          end
-       In module F:
+       In module "F":
        Modules do not match:
          functor (X : $S1) (Y : $S2) -> ...
        is not included in
@@ -2021,6 +2021,42 @@ Error: Signature mismatch:
             type 'a t = 'a * 'a
           is not included in
             type 'a t = 'a list
-          The type 'a * 'a is not equal to the type 'a list
+          The type "'a * 'a" is not equal to the type "'a list"
        2. Module types $S2 and $T2 match
 |}]
+
+
+(** Application of non-functor *)
+
+module K = List(A)(B)
+[%%expect {|
+Line 3, characters 11-21:
+3 | module K = List(A)(B)
+               ^^^^^^^^^^
+Error: The module "List" is not a functor, it cannot be applied.
+|}]
+
+module Error = (struct end)(B)
+[%%expect {|
+Line 1, characters 15-30:
+1 | module Error = (struct end)(B)
+                   ^^^^^^^^^^^^^^^
+Error: This module is not a functor, it cannot be applied.
+|}]
+
+let f (x:Set.Make(Set)(A).t) = x
+[%%expect {|
+Line 1, characters 9-27:
+1 | let f (x:Set.Make(Set)(A).t) = x
+             ^^^^^^^^^^^^^^^^^^
+Error: The functor application "Set.Make(Set)(A)" is ill-typed.
+       These arguments:
+         Set A
+       do not match these parameters:
+         functor (Ord : Set.OrderedType) -> ...
+       1. The following extra argument is provided Set : (module Set)
+       2. Modules do not match:
+            A : sig type a = A.a end
+          is not included in
+            Set.OrderedType
+|}]
