@@ -1,6 +1,6 @@
 (* TEST
-   flags = " -w +A "
-   * expect
+ flags = " -w +A ";
+ expect;
 *)
 
 module type U = sig end
